/** 
 * @file microdia-sysfs.c
 * @author Nicolas VIVIEN
 * @date 2008-02-01
 * @version v0.0.0
 *
 * @brief Driver for Microdia USB video camera
 *
 * @note Copyright (C) Nicolas VIVIEN
 *
 * @par Licences
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/kref.h>
#include <linux/device.h>

#include <linux/usb.h>
#include <media/v4l2-common.h>

#include "microdia.h"
#include "sn9c20x.h"


extern const struct microdia_coord microdia_image_sizes[MICRODIA_NBR_SIZES];


/** 
 * @brief show_release
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'release' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_release(struct class_device *class, char *buf)
#else
static ssize_t show_release(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->release);
}


/** 
 * @brief show_videostatus
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'videostatus' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_videostatus(struct class_device *class, char *buf)
#else
static ssize_t show_videostatus(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf,
			"Nbr ISOC errors    : %d\n"
			"Nbr dropped frames : %d\n"
			"Nbr dumped frames  : %d\n",
			dev->visoc_errors,
			dev->vframes_error,
			dev->vframes_dumped);
}


/** 
 * @brief show_informations
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'informations' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_informations(struct class_device *class, char *buf)
#else
static ssize_t show_informations(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	int width, height;
	char *pixelfmt = NULL;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	char *palette_rgb24 = "RGB24 - RGB-8-8-8 - 24 bits";
	char *palette_rgb32 = "RGB32 - RGB-8-8-8-8 - 32 bits";
	char *palette_bgr24 = "BGR24 - BGR-8-8-8 - 24 bits";
	char *palette_bgr32 = "BGR32 - BGR-8-8-8-8 - 32 bits";
	char *palette_uyvy = "UYVY - YUV 4:2:2 - 16 bits";
	char *palette_yuyv = "YUYV - YUV 4:2:2 - 16 bits";
	char *palette_i420 = "I420 - YUV 4:2:0 - 12 bits";


	switch (dev->vsettings.palette) {
		case MICRODIA_PALETTE_RGB24:
			pixelfmt = palette_rgb24;
			break;

		case MICRODIA_PALETTE_RGB32:
			pixelfmt = palette_rgb32;
			break;

		case MICRODIA_PALETTE_BGR24:
			pixelfmt = palette_bgr24;
			break;

		case MICRODIA_PALETTE_BGR32:
			pixelfmt = palette_bgr32;
			break;

		case MICRODIA_PALETTE_UYVY:
			pixelfmt = palette_uyvy;
			break;

		case MICRODIA_PALETTE_YUYV:
			pixelfmt = palette_yuyv;
			break;

		case MICRODIA_PALETTE_I420:
			pixelfmt = palette_i420;
			break;
	}

	switch (dev->resolution) {
		case MICRODIA_80x60:
		/*case MICRODIA_128x96:*/
		case MICRODIA_160x120:
		/*case MICRODIA_213x160: */
		case MICRODIA_320x240:
		case MICRODIA_640x480:
			width = microdia_image_sizes[MICRODIA_640x480].x;
			height = microdia_image_sizes[MICRODIA_640x480].y;
			break;

		case MICRODIA_800x600:
		case MICRODIA_1024x768:
		case MICRODIA_1280x1024:
			width = microdia_image_sizes[MICRODIA_1280x1024].x;
			height = microdia_image_sizes[MICRODIA_1280x1024].y;
			break;

		default:
			width = 0;
			height = 0;
	}

	return sprintf(buf,
			"Asked resolution  : %dx%d\n"
			"Driver resolution : %dx%d\n"
			"Webcam resolution : %dx%d\n"
			"\n"
			"%s\n"
			"\n"
			"Brightness        : 0x%X\n"
			"Contrast          : 0x%X\n"
			"Whiteness         : 0x%X\n"
			"Exposure          : 0x%X\n"
			"Sharpness         : 0x%X\n"
			"Horizontal flip   : %d\n"
			"Vertical flip     : %d\n"
			"Auto-exposure     : %d\n"
			"Auto-whitebalance : %d\n",
			dev->view.x, dev->view.y,
			microdia_image_sizes[dev->resolution].x, microdia_image_sizes[dev->resolution].y,
			width, height,
			pixelfmt,
			0xFFFF & dev->vsettings.brightness,
			0xFFFF & dev->vsettings.contrast,
			0xFFFF & dev->vsettings.whiteness,
			0xFFFF & dev->vsettings.exposure,
			0x3F & dev->vsettings.sharpness,
			dev->vsettings.hflip,
			dev->vsettings.vflip,
			dev->vsettings.auto_exposure,
			dev->vsettings.auto_whitebalance);
}


/** 
 * @brief show_fps
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'fps' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_fps(struct class_device *class, char *buf)
#else
static ssize_t show_fps(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->vsettings.fps);
}


/** 
 * @brief show_exposure
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'exposure' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_exposure(struct class_device *class, char *buf)
#else
static ssize_t show_exposure(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.exposure);
}

/** 
 * @brief store_exposure
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_exposure(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_exposure(struct device *class, struct device_attribute *attr, 
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.exposure = (int) value;

	dev_microdia_camera_set_exposure(dev);

	return strlen(buf);
}

/** 
 * @brief show_brightness
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'brightness' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_brightness(struct class_device *class, char *buf)
#else
static ssize_t show_brightness(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.brightness);
}

/** 
 * @brief store_brightness
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_brightness(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_brightness(struct device *class, struct device_attribute *attr, 
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.brightness = (int) value;

	dev_microdia_camera_set_brightness(dev);

	return strlen(buf);
}

/** 
 * @brief show_contrast
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'contrast' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_contrast(struct class_device *class, char *buf)
#else
static ssize_t show_contrast(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.contrast);
}


/** 
 * @brief store_contrast
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_contrast(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_contrast(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.contrast = (int) value;

	dev_microdia_camera_set_contrast(dev);
//	dev_microdia_set_camera_quality(dev);

	return strlen(buf);
}


/** 
 * @brief show_whitebalance
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'whitebalance' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_whitebalance(struct class_device *class, char *buf)
#else
static ssize_t show_whitebalance(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.whiteness);
}


/** 
 * @brief store_whitebalance
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_whitebalance(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_whitebalance(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.whiteness = (int) value;

	dev_microdia_camera_set_gamma(dev);
//	dev_microdia_set_camera_quality(dev);

	return strlen(buf);
}


/** 
 * @brief show_colour
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'colour' value
 * 
 * @returns Size of buffer
 */
/*#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_colour(struct class_device *class, char *buf)
#else
static ssize_t show_colour(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.colour);
}
*/

/** 
 * @brief store_colour
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
/*#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_colour(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_colour(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.colour = (int) value;

//	dev_microdia_set_camera_quality(dev);

	return strlen(buf);
}
*/


/** 
 * @brief show_sharpness
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'sharpness' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_sharpness(struct class_device *class, char *buf)
#else
static ssize_t show_sharpness(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%X\n", dev->vsettings.sharpness);
}

/** 
 * @brief store_sharpness
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_sharpness(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_sharpness(struct device *class, struct device_attribute *attr, 
		const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	if (value < 0 || value > 0x3f)
		return -EINVAL;

	dev->vsettings.sharpness = (int) value;

	dev_microdia_camera_set_sharpness(dev);

	return strlen(buf);
}


/** 
 * @brief show_hflip
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'hflip' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_hflip(struct class_device *class, char *buf)
#else
static ssize_t show_hflip(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->vsettings.hflip);
}


/** 
 * @brief store_hflip
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_hflip(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_hflip(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	if (strncmp(buf, "1", 1) == 0)
		dev->vsettings.hflip = 1;
	else if (strncmp(buf, "0", 1) == 0)
		dev->vsettings.hflip = 0;
	else
		return -EINVAL;

	dev_microdia_camera_set_hvflip(dev);

	return strlen(buf);
}


/** 
 * @brief show_vflip
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'vflip' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_vflip(struct class_device *class, char *buf)
#else
static ssize_t show_vflip(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->vsettings.vflip);
}


/** 
 * @brief store_vflip
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_vflip(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_vflip(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	if (strncmp(buf, "1", 1) == 0)
		dev->vsettings.vflip = 1;
	else if (strncmp(buf, "0", 1) == 0)
		dev->vsettings.vflip = 0;
	else
		return -EINVAL;

	dev_microdia_camera_set_hvflip(dev);

	return strlen(buf);
}


/** 
 * @brief show_autoexposure
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'hflip' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_autoexposure(struct class_device *class, char *buf)
#else
static ssize_t show_autoexposure(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->vsettings.auto_exposure);
}


/** 
 * @brief store_autoexposure
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_autoexposure(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_autoexposure(struct device *class, struct device_attribute *attr,
		const char *buf, size_t count)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	if (strncmp(buf, "1", 1) == 0)
		dev->vsettings.auto_exposure = 1;
	else if (strncmp(buf, "0", 1) == 0)
		dev->vsettings.auto_exposure = 0;
	else
		return -EINVAL;

	dev_microdia_camera_set_auto_exposure(dev);

	return strlen(buf);
}


/** 
 * @brief show_autowhitebalance
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'vflip' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_autowhitebalance(struct class_device *class, char *buf)
#else
static ssize_t show_autowhitebalance(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "%d\n", dev->vsettings.auto_whitebalance);
}


/** 
 * @brief store_autowhitebalance
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_autowhitebalance(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_autowhitebalance(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	if (strncmp(buf, "1", 1) == 0)
		dev->vsettings.auto_whitebalance = 1;
	else if (strncmp(buf, "0", 1) == 0)
		dev->vsettings.auto_whitebalance = 0;
	else
		return -EINVAL;

	dev_microdia_camera_set_auto_whitebalance(dev);

	return strlen(buf);
}


/** 
 * @brief show_register adress
 *
 * @param class Class device
 *
 * @retval buf Adress of buffer with the 'register' value
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_register(struct class_device *class, char *buf)
#else
static ssize_t show_register(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	__u8 buf1;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	usb_microdia_control_read(dev, dev->vsettings.reg, &buf1, 1);

	dev->vsettings.val = buf1;

	return sprintf(buf, "0x%x = 0x%x\n",dev->vsettings.reg, dev->vsettings.val);
}


/** 
 * @brief store_register
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_register(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_register(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;
	__u8 buf1;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	value = simple_strtoul(buf, &endp, 16);

	dev->vsettings.val = (int) value;
	buf1 = dev->vsettings.val;
	usb_microdia_control_write(dev, dev->vsettings.reg, &buf1, 1);
 
	return strlen(buf);
}

/** 
 * @brief show_adress
 *
 * @param class Class device
 *
 * @retval buf Buffer
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_adress(struct class_device *class, char *buf)
#else
static ssize_t show_adress(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);
 
	return sprintf(buf, "0x%x\n", dev->vsettings.reg);
}


/** 
 * @brief store_adress
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_adress(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_adress(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);
 
	value = simple_strtoul(buf, &endp, 16);
 
	dev->vsettings.reg = (int) value;

	return strlen(buf);
}

/** 
 * @brief show_i2cadr
 *
 * @param class Class device
 *
 * @retval buf Buffer
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_i2cadr(struct class_device *class, char *buf)
#else
static ssize_t show_i2cadr(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "0x%x\n", dev->vsettings.i2creg);
}
 
 
/** 
 * @brief store_i2cadr
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_i2cadr(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_i2cadr(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;
 
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);
 
	value = simple_strtoul(buf, &endp, 16);
 
	dev->vsettings.i2creg = (int) value;
 
	return strlen(buf);
}

/** 
 * @brief show_i2creg adress
 *
 * @param class Class device
 *
 * @retval buf Buffer
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_i2creg(struct class_device *class, char *buf)
#else
static ssize_t show_i2creg(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	__u8 buf1[8];
//	__u8 sensor_slave_id = 0x30; /* slave ID currently hard coded !! */
	__u8 i2c_flags = SN9C20X_I2C_2WIRE;
 
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	sn9c20x_read_i2c_data(dev, dev->vsettings.i2cslave /*sensor_slave_id*/,
			 dev->vsettings.reg, dev->vsettings.i2creg, i2c_flags, buf1);
 
	dev->vsettings.i2cval = buf1[0];
	return sprintf(buf, "I2C-register 0x%x = 0x%x 0x%x (%d)\n",
								dev->vsettings.i2creg,buf1[0],buf1[1],dev->vsettings.reg);
//	return strlen(buf);
}


/** 
 * @brief store_i2creg
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_i2creg(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_i2creg(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;

	__u8 buf1[5];
//	__u8 sensor_slave_id = 0x30; /* slave ID currently hard coded !! */
	__u8 i2c_flags = SN9C20X_I2C_2WIRE;
 
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);
 
	value = simple_strtoul(buf, &endp, 16);
 
	dev->vsettings.i2cval = (int) value;
	buf1[0] = ((dev->vsettings.i2cval & 0xff00) >> 8);
	buf1[1] = (dev->vsettings.i2cval & 0xFF);
	buf1[2] = 0;
	buf1[3] = 0;
	buf1[4] = 0;
 
	sn9c20x_write_i2c_data(dev,dev->vsettings.i2cslave /*sensor_slave_id*/, 
			dev->vsettings.reg, dev->vsettings.i2creg, i2c_flags, buf1);
//	sprintf(buf, "store_i2creg: set I2C-reg. 0x%x to 0x%x 0x%x (%d)\n", 
//						dev->vsettings.i2creg,buf1[0],buf1[1],dev->vsettings.reg);
	return strlen(buf);
}


/** 
 * @brief store_i2c_slave_adr
 *
 * @param class Class device
 * @param buf Buffer
 * @param count Counter
 *
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t store_i2c_slave_adr(struct class_device *class, const char *buf, size_t count)
#else
static ssize_t store_i2c_slave_adr(struct device *class, struct device_attribute *attr, const char *buf, size_t count)
#endif
{
	char *endp;
	unsigned long value;
 
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);
 
	value = simple_strtoul(buf, &endp, 16);
 
	dev->vsettings.i2cslave = (int) value;
 
	return strlen(buf);
}


/** 
 * @brief show_i2c_slave_adr
 *
 * @param class Class device
 *
 * @retval buf Buffer
 * 
 * @returns Size of buffer
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static ssize_t show_i2c_slave_adr(struct class_device *class, char *buf)
#else
static ssize_t show_i2c_slave_adr(struct device *class, struct device_attribute *attr, char *buf)
#endif
{
	struct video_device *vdev = to_video_device(class);
	struct usb_microdia *dev = video_get_drvdata(vdev);

	return sprintf(buf, "0x%x\n", dev->vsettings.i2cslave);
}




#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
static CLASS_DEVICE_ATTR(release, S_IRUGO, show_release, NULL);										/**< Release value */
static CLASS_DEVICE_ATTR(videostatus, S_IRUGO, show_videostatus, NULL);								/**< Video status */
static CLASS_DEVICE_ATTR(informations, S_IRUGO, show_informations, NULL);							/**< Informations */
static CLASS_DEVICE_ATTR(fps, S_IRUGO, show_fps, NULL);												/**< FPS value */
static CLASS_DEVICE_ATTR(brightness, S_IRUGO | S_IWUGO, show_brightness, store_brightness);			/**< Brightness value */
static CLASS_DEVICE_ATTR(exposure, S_IRUGO | S_IWUGO, show_exposure, store_exposure);			/**< Brightness exposure */
static CLASS_DEVICE_ATTR(contrast, S_IRUGO | S_IWUGO, show_contrast, store_contrast);				/**< Contrast value */
static CLASS_DEVICE_ATTR(whitebalance, S_IRUGO | S_IWUGO, show_whitebalance, store_whitebalance);	/**< Whitebalance value */
static CLASS_DEVICE_ATTR(sharpness, S_IRUGO | S_IWUGO, show_sharpness, store_sharpness);		/**< Sharpness value */
//static CLASS_DEVICE_ATTR(colour, S_IRUGO | S_IWUGO, show_colour, store_colour);						/**< Hue value */
static CLASS_DEVICE_ATTR(hflip, S_IRUGO | S_IWUGO, show_hflip, store_hflip);						/**< Horizontal flip value */
static CLASS_DEVICE_ATTR(vflip, S_IRUGO | S_IWUGO, show_vflip, store_vflip);						/**< Vertical flip value */
static CLASS_DEVICE_ATTR(auto_exposure, S_IRUGO | S_IWUGO, show_autoexposure, store_autoexposure);			/**< Automatic exposure control value */
static CLASS_DEVICE_ATTR(auto_whitebalance, S_IRUGO | S_IWUGO, show_autowhitebalance, store_autowhitebalance);		/**< Automatic whitebalance control value */
static CLASS_DEVICE_ATTR(adr, S_IRUGO | S_IWUGO, show_adress, store_adress);				/**< Register adress */
static CLASS_DEVICE_ATTR(reg, S_IRUGO | S_IWUGO, show_register, store_register);				/**< Register value */
static CLASS_DEVICE_ATTR(i2cadr, S_IRUGO | S_IWUGO, show_i2cadr, store_i2cadr);				/**< I2C Register adress */
static CLASS_DEVICE_ATTR(i2creg, S_IRUGO | S_IWUGO, show_i2creg, store_i2creg);				/**< I2C Register value */
static CLASS_DEVICE_ATTR(i2cslave, S_IRUGO | S_IWUGO, show_i2c_slave_adr, store_i2c_slave_adr);				/**< I2C Slave adress */
#else
static DEVICE_ATTR(release, S_IRUGO, show_release, NULL);											/**< Release value */
static DEVICE_ATTR(videostatus, S_IRUGO, show_videostatus, NULL);									/**< Video status */
static DEVICE_ATTR(informations, S_IRUGO, show_informations, NULL);									/**< Informations */
static DEVICE_ATTR(fps, S_IRUGO, show_fps, NULL);													/**< FPS value */
static DEVICE_ATTR(brightness, S_IRUGO | S_IWUGO, show_brightness, store_brightness);				/**< Brightness value */
static DEVICE_ATTR(exposure, S_IRUGO | S_IWUGO, show_exposure, store_exposure);				/**< Exposure value */
static DEVICE_ATTR(contrast, S_IRUGO | S_IWUGO, show_contrast, store_contrast);						/**< Contrast value */
static DEVICE_ATTR(whitebalance, S_IRUGO | S_IWUGO, show_whitebalance, store_whitebalance);			/**< Whitebalance value */
static DEVICE_ATTR(sharpness, S_IRUGO | S_IWUGO, show_sharpness, store_sharpness);			/**< Sharpness value */
//static DEVICE_ATTR(colour, S_IRUGO | S_IWUGO, show_colour, store_colour);							/**< Hue value */
static DEVICE_ATTR(hflip, S_IRUGO | S_IWUGO, show_hflip, store_hflip);							/**< Horizontal flip value */
static DEVICE_ATTR(vflip, S_IRUGO | S_IWUGO, show_vflip, store_vflip);							/**< Vertical flip value */
static DEVICE_ATTR(auto_exposure, S_IRUGO | S_IWUGO, show_autoexposure, store_autoexposure);				/**< Automatic exposure control value */
static DEVICE_ATTR(auto_whitebalance, S_IRUGO | S_IWUGO, show_autowhitebalance, store_autowhitebalance);		/**< Automatic whitebalance control value */
static DEVICE_ATTR(adr, S_IRUGO | S_IWUGO, show_adress, store_adress);						/**< Register adress */
static DEVICE_ATTR(reg, S_IRUGO | S_IWUGO, show_register, store_register);						/**< Register value */
static DEVICE_ATTR(i2cadr, S_IRUGO | S_IWUGO, show_i2cadr, store_i2cadr);						/**< I2C Register adress */
static DEVICE_ATTR(i2creg, S_IRUGO | S_IWUGO, show_i2creg, store_i2creg);						/**< I2C Register value */
static DEVICE_ATTR(i2cslave, S_IRUGO | S_IWUGO, show_i2c_slave_adr, store_i2c_slave_adr);						/**< I2C Slave adress */
#endif


/** 
 * @brief Create the 'sys' entries.
 *
 * This function permits to create all the entries in the 'sys' filesystem.
 *
 * @param vdev Video device structure
 * 
 * @returns 0 if all is OK
 */
int microdia_create_sysfs_files(struct video_device *vdev)
{
	int ret;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	ret = video_device_create_file(vdev, &class_device_attr_release);
	ret = video_device_create_file(vdev, &class_device_attr_videostatus);
	ret = video_device_create_file(vdev, &class_device_attr_informations);
	ret = video_device_create_file(vdev, &class_device_attr_fps);
	ret = video_device_create_file(vdev, &class_device_attr_brightness);
	ret = video_device_create_file(vdev, &class_device_attr_exposure);
	ret = video_device_create_file(vdev, &class_device_attr_contrast);
	ret = video_device_create_file(vdev, &class_device_attr_whitebalance);
	ret = video_device_create_file(vdev, &class_device_attr_sharpness);
	/*ret = video_device_create_file(vdev, &class_device_attr_colour);*/
	ret = video_device_create_file(vdev, &class_device_attr_hflip);
	ret = video_device_create_file(vdev, &class_device_attr_vflip);
	ret = video_device_create_file(vdev, &class_device_attr_auto_exposure);
	ret = video_device_create_file(vdev, &class_device_attr_auto_whitebalance);
	ret = video_device_create_file(vdev, &class_device_attr_reg);
	ret = video_device_create_file(vdev, &class_device_attr_adr);
	ret = video_device_create_file(vdev, &class_device_attr_i2creg);
	ret = video_device_create_file(vdev, &class_device_attr_i2cadr);
	ret = video_device_create_file(vdev, &class_device_attr_i2cslave);
#else
	ret = video_device_create_file(vdev, &dev_attr_release);
	ret = video_device_create_file(vdev, &dev_attr_videostatus);
	ret = video_device_create_file(vdev, &dev_attr_informations);
	ret = video_device_create_file(vdev, &dev_attr_fps);
	ret = video_device_create_file(vdev, &dev_attr_brightness);
	ret = video_device_create_file(vdev, &dev_attr_exposure);
	ret = video_device_create_file(vdev, &dev_attr_contrast);
	ret = video_device_create_file(vdev, &dev_attr_whitebalance);
	ret = video_device_create_file(vdev, &dev_attr_sharpness);
	/*ret = video_device_create_file(vdev, &dev_attr_colour);*/
	ret = video_device_create_file(vdev, &dev_attr_hflip);
	ret = video_device_create_file(vdev, &dev_attr_vflip);
	ret = video_device_create_file(vdev, &dev_attr_auto_exposure);
	ret = video_device_create_file(vdev, &dev_attr_auto_whitebalance);
	ret = video_device_create_file(vdev, &dev_attr_reg);
	ret = video_device_create_file(vdev, &dev_attr_adr);
	ret = video_device_create_file(vdev, &dev_attr_i2creg);
	ret = video_device_create_file(vdev, &dev_attr_i2cadr);
	ret = video_device_create_file(vdev, &dev_attr_i2cslave);
#endif
	return ret;
}


/** 
 * @brief Remove the 'sys' entries.
 *
 * This function permits to remove all the entries in the 'sys' filesystem.
 *
 * @param vdev Video device structure
 * 
 * @returns 0 if all is OK
 */
void microdia_remove_sysfs_files(struct video_device *vdev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	video_device_remove_file(vdev, &class_device_attr_release);
	video_device_remove_file(vdev, &class_device_attr_videostatus);
	video_device_remove_file(vdev, &class_device_attr_informations);
	video_device_remove_file(vdev, &class_device_attr_fps);
	video_device_remove_file(vdev, &class_device_attr_brightness);
	video_device_remove_file(vdev, &class_device_attr_exposure);
	video_device_remove_file(vdev, &class_device_attr_contrast);
	video_device_remove_file(vdev, &class_device_attr_whitebalance);
	video_device_remove_file(vdev, &class_device_attr_sharpness);
	/*video_device_remove_file(vdev, &class_device_attr_colour);*/
	video_device_remove_file(vdev, &class_device_attr_hflip);
	video_device_remove_file(vdev, &class_device_attr_vflip);
	video_device_remove_file(vdev, &class_device_attr_auto_exposure);
	video_device_remove_file(vdev, &class_device_attr_auto_whitebalance);
	video_device_remove_file(vdev, &class_device_attr_reg);
	video_device_remove_file(vdev, &class_device_attr_adr);
	video_device_remove_file(vdev, &class_device_attr_i2creg);
	video_device_remove_file(vdev, &class_device_attr_i2cadr);
	video_device_remove_file(vdev, &class_device_attr_i2cslave);
#else
	video_device_remove_file(vdev, &dev_attr_release);
	video_device_remove_file(vdev, &dev_attr_videostatus);
	video_device_remove_file(vdev, &dev_attr_informations);
	video_device_remove_file(vdev, &dev_attr_fps);
	video_device_remove_file(vdev, &dev_attr_brightness);
	video_device_remove_file(vdev, &dev_attr_exposure);
	video_device_remove_file(vdev, &dev_attr_contrast);
	video_device_remove_file(vdev, &dev_attr_whitebalance);
	video_device_remove_file(vdev, &dev_attr_sharpness);
	/*video_device_remove_file(vdev, &dev_attr_colour);*/
	video_device_remove_file(vdev, &dev_attr_hflip);
	video_device_remove_file(vdev, &dev_attr_vflip);
	video_device_remove_file(vdev, &dev_attr_auto_exposure);
	video_device_remove_file(vdev, &dev_attr_auto_whitebalance);
	video_device_remove_file(vdev, &dev_attr_reg);
	video_device_remove_file(vdev, &dev_attr_adr);
	video_device_remove_file(vdev, &dev_attr_i2creg);
	video_device_remove_file(vdev, &dev_attr_i2cadr);
	video_device_remove_file(vdev, &dev_attr_i2cslave);
#endif
}

