/** 
 * @file microdia-usb.c
 * @author Nicolas VIVIEN
 * @date 2008-02-01
 * @version v0.0.0
 *
 * @brief Driver for Microdia USB video camera
 *
 * @note Copyright (C) Nicolas VIVIEN
 *
 * @par Licences
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/kref.h>
#include <linux/debugfs.h>
#include <linux/stat.h>

#include <linux/usb.h>
#include <media/v4l2-common.h>

#include "microdia.h"
#include "sn9c20x.h"
#include "mt9vx11.h"

/**
 * @var fps
 *   Module parameter to set frame per second
 */
static int fps = 25;

/**
 * @var hflip
 *  Module parameter to enable/disable the horizontal flip process
 */
static int hflip = 0;

/**
 * @var flip_detect
 *  Module parameter to enable/disable vflip detection
 */
static int flip_detect = 0;

/**
 * @var flip_detect
 *  Module parameter to enable/disable hflip detect
 */

/**
 * @var vflip
 *   Module parameter to enable/disable the vertical flip process
 */
static int vflip = 0;

/**
 * @var brightness
 *   Module parameter to set the brightness
 */
static int brightness = MICRODIA_PERCENT(50, 0xFFFF);

/**
 * @var whiteness
 *   Module parameter to set the whiteness
 */
static int whiteness = MICRODIA_PERCENT(20, 0xFFFF);

/**
 * @var contrast
 *   Module parameter to set the contrast
 */
static int contrast = MICRODIA_PERCENT(50, 0xFFFF);

/**
 * @var exposure
 *   Module parameter to set the exposure
 */
static int exposure = MICRODIA_PERCENT(20, 0xFFFF);

/**
 * @var sharpness
 *   Module parameter to set the exposure
 */
static int sharpness = MICRODIA_PERCENT(50, 0x3F);

/**
 * @var auto_exposure
 *   Module parameter to set the exposure
 */
static int auto_exposure = 1;

/**
 * @var auto_whitebalance
 *   Module parameter to set the exposure
 */
static int auto_whitebalance = 1;

/**
 * @var debug_dir_name
 *   Name of our directory in debugfs
 */
static const char * debug_dir_name = "microdia";

/**
 * @var debug_file_name
 *   Name of our debug file in debugfs
 */
static const char * debug_file_name = "debug_level";

/**
 * @var debug_dir
 *   Dentry for our debug dir (for cleanup)
 */
static struct dentry * debug_dir;

/**
 * @var debug_file
 *   Dentry for our debug file (for cleanup)
 */
static struct dentry * debug_file;

/**
 * @var debug_level
 *   Level at which we do debugging
 */
static __u8 debug_level = 0;


/**
 * @var microdia_table
 * Define all the hotplug supported devices by this driver
 */
static struct usb_device_id microdia_table[] = {
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6240) },	// SN9C201 + MI1300
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6242) },	// SN9C201 + MI1310
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6243) },	// SN9C201 + S5K4AAFX
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6248) },	// SN9C201 + OV9655
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_624B) },	// SN9C201 + CX1332
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_624C) },	// SN9C201 + MI1320
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_624E) },	// SN9C201 + SOI968
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_624F) },	// SN9C201 + OV9650
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6253) },	// SN9C201 + OV9650
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6260) },	// SN9C201 + OV7670ISP
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6262) },	// SN9C201 + OM6802
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6270) },	// SN9C201 + MI0360/MT9V111
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_627A) },	// SN9C201 + S5K53BEB
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_627B) },	// SN9C201 + OV7660
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_627C) },	// SN9C201 + HV7131R
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_627F) },	// EEPROM
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6280) },	// SN9C202 + MI1300
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6282) },	// SN9C202 + MI1310
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6283) },	// SN9C202 + S5K4AAFX
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6288) },	// SN9C202 + OV9655
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_628A) },	// SN9C202 + ICM107
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_628B) },	// SN9C202 + CX1332
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_628C) },	// SN9C202 + MI1320
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_628E) },	// SN9C202 + SOI968
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_628F) },	// SN9C202 + OV9650
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_62A0) },	// SN9C202 + OV7670ISP
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_62A2) },	// SN9C202 + OM6802
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_62B0) },	// SN9C202 + MI0360/MT9V111
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_62B3) },	// SN9C202 + OV9655
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_62BA) },	// SN9C202 + S5K53BEB
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_62BB) },	// SN9C202 + OV7660
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_62BC) },	// SN9C202 + HV7131R
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_62BE) },	// SN9C202 + OV7663
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_6128) },	// SN9C325 + OM6802
	{ USB_DEVICE(USB_VENDOR_ID_MICRODIA, USB_UDIA_MICRODIA_PRODUCT_ID_612A) },	// SN9C325 + OV7648 + POx1030xC + SOI768 + PO2030N + OV7660 + OV7670 + HV7131R
	{ USB_DEVICE(USB_VENDOR_ID_MICROSOFT, USB_UDIA_MICROSOFT_PRODUCT_ID_00F4) },	// => 628f (SN9C202 + OV9650)
	{ USB_DEVICE(USB_VENDOR_ID_TRUST, USB_UDIA_TRUST_PRODUCT_ID_013D) },		// => 627b (SN9C201 + OV7660)
	{ USB_DEVICE(USB_VENDOR_ID_CHICONY, USB_UDIA_CHICONY_PRODUCT_ID_A128) },	// => 62be (SN9C202 + OV7663 + EEPROM)
	{ }
};


MODULE_DEVICE_TABLE(usb, microdia_table);		/**< Define the supported devices */



/** 
 * @param dev Device structure
 * 
 * @returns 0 if all is OK
 *
 * @brief Initilize an isochronous pipe.
 *
 * This function permits to initialize an URB transfert (or isochronous pipe).
 */
int usb_microdia_isoc_init(struct usb_microdia *dev)
{
	int i, j;
	int ret = 0;
	__u16 iso_max_frame_size;
	struct urb *urb;
	struct usb_device *udev;

	if (dev == NULL)
		return -EFAULT;

	if (dev->isoc_init_ok)
		return 0;

	udev = dev->udev;

	UDIA_DEBUG("usb_microdia_isoc_init()\n");

	// Allocate URB structure
	for (i=0; i<MAX_ISO_BUFS; i++) {
		urb = usb_alloc_urb(ISO_FRAMES_PER_DESC, GFP_KERNEL);

		if (urb == NULL) {
			UDIA_ERROR("Failed to allocate URB %d\n", i);
			ret = -ENOMEM;
			break;
		}

		dev->isobuf[i].urb = urb;
	}

	if (ret) {
		while (i >= 0) {
			if (dev->isobuf[i].urb != NULL)
				usb_free_urb(dev->isobuf[i].urb);

			dev->isobuf[i].urb = NULL;
			i--;
		}

		return ret;
	}

  
  // iso_max_frame_size should be the endpoints maximum packetsize * the highbandwidth multiplier.

	iso_max_frame_size = max_packet_sz(le16_to_cpu(dev->isoc_in_size)) * hb_multiplier(le16_to_cpu(dev->isoc_in_size));
	
	// Init URB structure
	for (i=0; i<MAX_ISO_BUFS; i++) {
		urb = dev->isobuf[i].urb;

		urb->interval = 1; 
		urb->dev = udev;
		urb->pipe = usb_rcvisocpipe(udev, dev->isoc_in_endpointAddr);
		urb->transfer_flags = URB_ISO_ASAP;
		urb->transfer_buffer = dev->isobuf[i].data;
		urb->transfer_buffer_length = iso_max_frame_size * ISO_FRAMES_PER_DESC;
		urb->complete = usb_microdia_isoc_handler;
		urb->context = dev;
		urb->start_frame = 0;
		urb->number_of_packets = ISO_FRAMES_PER_DESC;

		for (j=0; j<ISO_FRAMES_PER_DESC; j++) {
			urb->iso_frame_desc[j].offset = j * iso_max_frame_size;
			urb->iso_frame_desc[j].length = iso_max_frame_size; //dev->isoc_in_size;
		}
	}

	UDIA_DEBUG("dev->isoc_in_size = %X\n", dev->isoc_in_size);
	UDIA_DEBUG("dev->isoc_in_endpointAddr = %X\n", dev->isoc_in_endpointAddr);

	// Link
	for (i=0; i<MAX_ISO_BUFS; i++) {
		ret = usb_submit_urb(dev->isobuf[i].urb, GFP_KERNEL);

		if (ret)
			UDIA_ERROR("isoc_init() submit_urb %d failed with error %d\n", i, ret);
		else
			UDIA_DEBUG("URB 0x%p submitted.\n", dev->isobuf[i].urb);

		switch (ret) {
			case -ENOMEM:
				UDIA_ERROR("ENOMEM\n");
				break;
			case -ENODEV:
				UDIA_ERROR("ENODEV\n");
				break;
			case -ENXIO:
				UDIA_ERROR("ENXIO\n");
				break;
			case -EINVAL:
				UDIA_ERROR("EINVAL\n");
				break;
			case -EAGAIN:
				UDIA_ERROR("EAGAIN\n");
				break;
			case -EFBIG:
				UDIA_ERROR("EFBIG\n");
				break;
			case -EPIPE:
				UDIA_ERROR("EPIPE\n");
				break;
			case -EMSGSIZE:
				UDIA_ERROR("EMSGSIZE\n");
				break;
		}
	}

	// All is done
	dev->isoc_init_ok = 1;

	return 0;
}


/** 
 * @param urb URB structure
 *
 * @brief ISOC handler
 *
 * This function is called as an URB transfert is complete (Isochronous pipe).
 * So, the traitement is done in interrupt time, so it has be fast, not crash,
 * ans not stall. Neat.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
void usb_microdia_isoc_handler(struct urb *urb, struct pt_regs *regs)
#else
void usb_microdia_isoc_handler(struct urb *urb)
#endif
{
	int i;
	int ret;
	int skip;

	int awake = 0;
	int framestatus;
	int framelen;

	unsigned char *fill = NULL;
	unsigned char *iso_buf = NULL;

	struct usb_microdia *dev;
	struct microdia_frame_buf *framebuf;

	unsigned char frame_header[] = {0xff, 0xff, 0x00, 0xc4, 0xc4, 0x96};

	UDIA_STREAM("Isoc handler\n");

	dev = (struct usb_microdia *) urb->context;

	if (dev == NULL) {
		UDIA_ERROR("isoc_handler called with NULL device !\n");
		return;
	}

	if (urb->status == -ENOENT || urb->status == -ECONNRESET) {
		UDIA_DEBUG("URB unlinked synchronuously !\n");
		return;
	}

	if (urb->status != -EINPROGRESS && urb->status != 0) {
		const char *errmsg;

		errmsg = "Unknown";

		switch(urb->status) {
			case -ENOSR:
				errmsg = "Buffer error (overrun)";
				break;

			case -EPIPE:
				errmsg = "Stalled (device not responding)";
				break;

			case -EOVERFLOW:
				errmsg = "Babble (bad cable?)";
				break;

			case -EPROTO:
				errmsg = "Bit-stuff error (bad cable?)";
				break;

			case -EILSEQ:
				errmsg = "CRC/Timeout (could be anything)";
				break;

			case -ETIMEDOUT:
				errmsg = "NAK (device does not respond)";
				break;
		}

		UDIA_ERROR("isoc_handler() called with status %d [%s].\n", urb->status, errmsg);

		dev->visoc_errors++;

		wake_up_interruptible(&dev->wait_frame);

		urb->dev = dev->udev;
		ret = usb_submit_urb(urb, GFP_ATOMIC);

		if (ret != 0) {
			UDIA_ERROR("Error (%d) re-submitting urb in microdia_isoc_handler.\n", ret);
		}

		return;
	}

	framebuf = dev->fill_frame;

	if (framebuf == NULL) {
		UDIA_ERROR("isoc_handler without valid fill frame !\n");
		
		wake_up_interruptible(&dev->wait_frame);

		urb->dev = dev->udev;
		ret = usb_submit_urb(urb, GFP_ATOMIC);

		if (ret != 0) {
			UDIA_ERROR("Error (%d) re-submitting urb in microdia_isoc_handler.\n", ret);
		}

		return;
	}
	else {
		fill = framebuf->data + framebuf->filled;
	}

	// Reset ISOC error counter
	dev->visoc_errors = 0;

	// Compact data
	for (i=0; i<urb->number_of_packets; i++) {
		framestatus = urb->iso_frame_desc[i].status;
		framelen = urb->iso_frame_desc[i].actual_length;
		iso_buf = urb->transfer_buffer + urb->iso_frame_desc[i].offset;

		if (framestatus == 0) {
			// By default, I keep all packet.
			// So I don't skip byte !
			skip = 0;

			if (framelen > 0) {
				// For microdia devices, data are isoc packet with site > 0.
				// ...
				if (memcmp(iso_buf, frame_header, 6) == 0) {
					/*
					 * Correct frame_size is set in
					 * v4l_microdia_select_video_mode
					 */
					if ((framebuf->filled ==
							dev->frame_size) &&
							framebuf->errors == 0) {
						if (microdia_next_frame(dev)) {
							dev->vframes_dumped++;
						}
					} else {
						dev->vframes_error++;
					}
					awake = 1;
					framebuf = dev->fill_frame;
					framebuf->filled = 0;
					framebuf->errors = 0;
					fill = framebuf->data;
					skip = 64;
				}
				// Our buffer is full !!!
				if (framelen - skip + framebuf->filled >
						dev->frame_size) {
					UDIA_ERROR("Frame buffer overflow !\n");
					framebuf->errors++;
				}
				// All is OK
				else {
					memcpy(fill, iso_buf + skip,
							framelen - skip);
					fill += framelen - skip;
				}

				// New size of our buffer
				framebuf->filled += framelen - skip;
			}

			UDIA_STREAM("URB : Length = %d - Skip = %d - Buffer size = %d\n",
				framelen, skip, framebuf->filled);
		}
		else {
			UDIA_ERROR("Iso frame %d of USB has error %d\n", i, framestatus);
		}
	}

	if (awake == 1)
		wake_up_interruptible(&dev->wait_frame);

	urb->dev = dev->udev;

	ret = usb_submit_urb(urb, GFP_ATOMIC);

	if (ret != 0) {
		UDIA_ERROR("Error (%d) re-submitting urb in microdia_isoc_handler.\n", ret);
	}
}


/** 
 * @param dev Device structure
 *
 * @brief Clean-up all the ISOC buffers
 *
 * This function permits to clean-up all the ISOC buffers.
 */
void usb_microdia_isoc_cleanup(struct usb_microdia *dev)
{
	int i;

	UDIA_DEBUG("Isoc cleanup\n");

	if (dev == NULL)
		return;

	if (dev->isoc_init_ok == 0)
		return;

	// Unlinking ISOC buffers
	for (i=0; i<MAX_ISO_BUFS; i++) {
		struct urb *urb;

		urb = dev->isobuf[i].urb;

		if (urb != 0) {
			if (dev->isoc_init_ok)
				usb_kill_urb(urb);
			
			usb_free_urb(urb);
			dev->isobuf[i].urb = NULL;
		}
	}

	// All is done
	dev->isoc_init_ok = 0;
}



/** 
 * @param dev Device structure
 * @param index Choice of the interface
 * 
 * @returns 0 if all is OK
 *
 * @brief Send the message SET_FEATURE and choose the interface
 *
 * This function permits to send the message SET_FEATURE on the USB bus.
 */
int usb_microdia_set_feature(struct usb_microdia *dev, int index)
{
	int result;
	struct usb_device *udev = dev->udev;

	result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
			USB_REQ_SET_FEATURE,
			USB_TYPE_STANDARD | USB_DIR_OUT | USB_RECIP_DEVICE,
			USB_DEVICE_REMOTE_WAKEUP,
			index,
			NULL,
			0,
			500);
	
	if (result < 0)
		UDIA_ERROR("SET FEATURE fail !\n");
	else 
		UDIA_DEBUG("SET FEATURE\n");

	return result;
}


/** 
 * @param dev Device structure
 * 
 * @returns 0 if all is OK
 *
 * @brief Send the message SET_CONFIGURATION
 *
 * This function permits to send the message SET_CONFIGURATION on the USB bus.
 */
int usb_microdia_set_configuration(struct usb_microdia *dev)
{
	int result;
	struct usb_device *udev = dev->udev;

	result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
			USB_REQ_SET_CONFIGURATION,
			USB_TYPE_STANDARD | USB_DIR_OUT | USB_RECIP_DEVICE,
			0,
			udev->config[0].desc.bConfigurationValue,
			NULL,
			0,
			500);

	if (result < 0)
		UDIA_ERROR("SET CONFIGURATION fail !\n");
	else 
		UDIA_DEBUG("SET CONFIGURATION %d\n", udev->config[0].desc.bConfigurationValue);

	return result;
}


/**
 * @param dev Device structure
 * @param value register to write to
 * @param data
 * @param length number of bytes
 *
 * @returns 0 if all is OK
 *
 * @brief Write a 16-bit value to a 16-bit register
 *
 * This function permits to write a 16-bit value to a 16-bit register on the USB bus.
 */
int usb_microdia_control_write(struct usb_microdia *dev, __u16 value, __u8 *data, __u16 length)
{
	int result;
	struct usb_device *udev = dev->udev;

	result = usb_control_msg(udev, usb_sndctrlpipe(udev, 0),
			0x08,
			USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_INTERFACE,
			value,
			0x00,
			data,
			length,
			500);

	if (result < 0)
		UDIA_ERROR("Write register failed index = 0x%02X\n", value);

	return result;
}

/**
 * @param dev
 * @param commands
 * @param data
 * @param cmdlen
 * @param datalen
 *
 * @returns 0 if all is OK
 *
 * @brief Write a series of 16-bit "commands" using the same buffer
 */
int usb_microdia_control_write_multi(struct usb_microdia *dev, __u16 *commands, __u8 *data,
	__u16 cmdlen, __u16 datalen)
{
	int result, i;

	for(i = 0; i < cmdlen; i++) {
		result = usb_microdia_control_write(dev, commands[i], data, datalen);

		if(result < 0)
			return result;
	}

	return 0;
}

/**
 * @param dev Device structure
 * @param index register to read from
 * @param data
 * @param length number of bytes
 *
 * @returns 0 if all is OK
 *
 * @brief Read a 16-bit value from a 16-bit register
 *
 * This function permits to read a 16-bit value from a 16-bit register on the USB bus.
 */
int usb_microdia_control_read(struct usb_microdia *dev, __u16 index, __u8 *data, __u16 length)
{
	int result;

	struct usb_device *udev = dev->udev;

	*data = 0;

	result = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
			0x00,
			USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_INTERFACE,
			index,
			0x00,
			data,
			length,
			500);

	if (result < 0)
		UDIA_ERROR("Read register failed 0x%02X\n", index);

	return result;
}


/** 
 * @param dev 
 * 
 * @returns 0 if all is OK
 *
 * @brief Set the default value about the video settings.
 *
 * This function permits to set the video settings for each video camera model.
 * 
 */
static int usb_microdia_default_settings(struct usb_microdia *dev)
{
	switch (dev->webcam_model) {
		default:
			dev->vsettings.fps = fps;
			dev->vsettings.vflip = vflip;
			dev->vsettings.hflip = hflip;
			dev->vsettings.brightness = brightness & 0xffff;
			dev->vsettings.contrast = contrast & 0xffff;
			dev->vsettings.whiteness = whiteness & 0xffff;
			dev->vsettings.exposure = exposure & 0xffff;
			dev->vsettings.sharpness = sharpness & 0x3f;
			dev->vsettings.auto_exposure = auto_exposure & 1;
			dev->vsettings.auto_whitebalance = auto_whitebalance & 1;
			break;
	}

	return 0;
}

extern int microdia_6242_start_stream(struct usb_microdia *dev);
extern int microdia_6242_stop_stream(struct usb_microdia *dev);
extern int microdia_624e_initialize(struct usb_microdia *dev);
extern int microdia_624e_start_stream(struct usb_microdia *dev);
extern int microdia_624e_stop_stream(struct usb_microdia *dev);
extern int microdia_624f_initialize(struct usb_microdia *dev);
extern int microdia_624f_stop_stream(struct usb_microdia *dev);
extern int microdia_624f_start_stream(struct usb_microdia *dev);
extern int microdia_624f_set_exposure(struct usb_microdia *dev);
extern int microdia_624f_flip_detect(struct usb_microdia *dev);
extern int microdia_6260_initialize(struct usb_microdia *dev);
extern int microdia_6260_start_stream(struct usb_microdia *dev);
extern int microdia_6260_stop_stream(struct usb_microdia *dev);
extern int microdia_6260_flip_detect(struct usb_microdia *dev);
extern int microdia_6270_initialize(struct usb_microdia *dev);
extern int microdia_6270_start_stream(struct usb_microdia *dev);
extern int microdia_6270_stop_stream(struct usb_microdia *dev);
extern int microdia_627b_initialize(struct usb_microdia *dev);
extern int microdia_627b_start_stream(struct usb_microdia *dev);
extern int microdia_627b_stop_stream(struct usb_microdia *dev);
extern int microdia_627f_initialize(struct usb_microdia *dev);
extern int microdia_6288_initialize(struct usb_microdia *dev);
extern int microdia_6288_start_stream(struct usb_microdia *dev);
extern int microdia_6288_stop_stream(struct usb_microdia *dev);
extern int microdia_6128_initialize(struct usb_microdia *dev);
extern int microdia_6128_start_stream(struct usb_microdia *dev);
extern int microdia_6128_stop_stream(struct usb_microdia *dev);


/** 
 * @param interface 
 * @param id 
 * 
 * @returns 0 if all is OK
 *
 * @brief Load the driver
 *
 * This function detects the device and allocate the buffers for the device
 * and the video interface.
 */
static int usb_microdia_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	int i;
	int err;
	size_t buffer_size;

	int vendor_id;
	int product_id;
	int bNumInterfaces;
	int webcam_model;
	int webcam_type;
	__u8 sensor_slave_address = 0;
	__u8 sensor_flags = 0x80;	// 2-WIRE,100KHz
	int (*initialize)(struct usb_microdia *) = NULL;
	int (*stop_stream)(struct usb_microdia *) = NULL;
	int (*start_stream)(struct usb_microdia *) = NULL;
	int (*set_contrast)(struct usb_microdia *) = NULL;
	int (*set_brightness)(struct usb_microdia *) = NULL;
	int (*set_gamma)(struct usb_microdia *) = NULL;
	int (*set_exposure)(struct usb_microdia *) = NULL;
	int (*flip_detect_f)(struct usb_microdia *) = NULL;
	int (*set_hvflip)(struct usb_microdia *) = NULL;
	int (*set_sharpness)(struct usb_microdia *) = NULL;
	int (*set_auto_exposure)(struct usb_microdia *) = NULL;
	int (*set_auto_whitebalance)(struct usb_microdia *) = NULL;

	struct usb_microdia *dev = NULL;
	struct usb_device *udev = interface_to_usbdev(interface);
	struct usb_host_interface *iface_desc;
	struct usb_endpoint_descriptor *endpoint;

	// Set default device parameters:
	webcam_type = MICRODIA_VGA;
	sensor_flags = SN9C20X_I2C_2WIRE;
	set_contrast = sn9c20x_set_contrast;
	set_brightness = sn9c20x_set_brightness;
	set_gamma = sn9c20x_set_gamma;
	set_sharpness = sn9c20x_set_sharpness;

	// Get USB VendorID and ProductID
	vendor_id = le16_to_cpu(udev->descriptor.idVendor);
	product_id = le16_to_cpu(udev->descriptor.idProduct);

	// Check if we can handle this device
	UDIA_DEBUG("Probe function called with VendorID=%04X, ProductID=%04X and InterfaceNumber=%d\n",
			vendor_id, product_id, interface->cur_altsetting->desc.bInterfaceNumber);

	// The interface are probed one by one.
	// We are interested in the video interface (always the interface '0')
	// The interfaces '1' or '2' (if presents) are the audio control.
	if (interface->cur_altsetting->desc.bInterfaceNumber > 0)
		return -ENODEV;

	// Detect device
	if (vendor_id == USB_VENDOR_ID_MICRODIA) {
		switch (product_id) {
			case USB_UDIA_MICRODIA_PRODUCT_ID_6240:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6240.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_6240;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6242:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6242.\n");
				webcam_model = MICRODIA_6242;
				sensor_slave_address = 0x5d;
				initialize = microdia_624e_initialize;
				start_stream = microdia_6242_start_stream;
				stop_stream = microdia_6242_stop_stream;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6243:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6243.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_6243;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6248:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6248.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_6248;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_624B:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 624B.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_624B;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_624C:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 624C.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_624C;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_624E:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 624E.\n");
				webcam_model = MICRODIA_624E;
				webcam_type = MICRODIA_SXGA;
				sensor_slave_address = 0x30;
				initialize = microdia_624e_initialize;
				start_stream = microdia_624e_start_stream;
				stop_stream = microdia_624e_stop_stream;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_624F:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 624F.\n");
				webcam_model = MICRODIA_624F;
				webcam_type = MICRODIA_SXGA;
				sensor_slave_address = 0x30;
				initialize = microdia_624f_initialize;
				start_stream = microdia_624f_start_stream;
				stop_stream = microdia_624f_stop_stream;
				set_exposure = microdia_624f_set_exposure;
				flip_detect_f = microdia_624f_flip_detect;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6253:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6253.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_6253;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6260:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6260.\n");
				webcam_model = MICRODIA_6260;
				sensor_slave_address = 0x21;
				initialize = microdia_6260_initialize;
				start_stream = microdia_6260_start_stream;
				stop_stream = microdia_6260_stop_stream;
				flip_detect_f = microdia_6260_flip_detect;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6262:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6262.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_6262;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6270:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6270.\n");
				webcam_model = MICRODIA_6270;
				sensor_slave_address = 0;	// will be probed during initialisation
				initialize = microdia_6270_initialize;
				start_stream = microdia_6270_start_stream;
				stop_stream = microdia_6270_stop_stream;
				set_exposure = mt9vx11_set_exposure;
				set_hvflip = mt9vx11_set_hvflip;
				set_auto_exposure = mt9v111_set_autoexposure;
				set_auto_whitebalance = mt9v111_set_autowhitebalance;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_627A:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 627A.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_627A;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_627B:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 627B.\n");
				webcam_model = MICRODIA_627B;
				sensor_slave_address = 0x21;
				initialize = microdia_627b_initialize;
				start_stream = microdia_627b_start_stream;
				stop_stream = microdia_627b_stop_stream;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_627C:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 627C.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_627C;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_627F:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 627F.\n");
				webcam_model = MICRODIA_627F;
				webcam_type = MICRODIA_SXGA;
				sensor_slave_address = 0x30;
				initialize = microdia_627f_initialize;
				start_stream = microdia_624f_start_stream;
				stop_stream = microdia_624f_stop_stream;
				set_exposure = microdia_624f_set_exposure;
				//flip_detect_f = microdia_624f_flip_detect;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6280:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6280.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_6280;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6282:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6282.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_6282;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6283:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6283.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_6283;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6288:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6288.\n");
				webcam_model = MICRODIA_6288;
				sensor_slave_address = 0x30;
				initialize = microdia_6288_initialize;
				start_stream = microdia_6288_start_stream;
				stop_stream = microdia_6288_stop_stream;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_628A:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 628A.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_628A;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_628B:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 628B.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_628B;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_628C:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 628C.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_628C;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_628E:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 628E.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_628E;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_628F:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 628F.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_628F;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_62A0:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 62A0.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_62A0;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_62A2:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 62A2.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_62A2;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_62B0:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 62B0.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_62B0;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_62B3:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 62B3.\n");
				webcam_model = MICRODIA_62B3;
				sensor_slave_address = 0x30;
				initialize = microdia_6288_initialize;
				start_stream = microdia_6288_start_stream;
				stop_stream = microdia_6288_stop_stream;
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_62BA:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 62BA.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_62BA;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_62BB:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 62BB.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_62BB;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_62BC:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 62BC.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_62BC;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_62BE:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 62BE.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_62BE;
				return -ENODEV;

			case USB_UDIA_MICRODIA_PRODUCT_ID_6128:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 6128.\n");
				webcam_model = MICRODIA_6128;
				webcam_type = MICRODIA_SXGA;
				sensor_slave_address = 0x34;
				initialize = microdia_6128_initialize;
				start_stream = microdia_6128_start_stream;
				stop_stream = microdia_6128_stop_stream;
				set_contrast = NULL;	// sn9c325_set_contrast
				set_brightness = NULL;	// sn9c325_set_brightness
				set_gamma = NULL;	// sn9c325_set_gamma
				set_sharpness = NULL;	// sn9c325_set_sharpness
				break;

			case USB_UDIA_MICRODIA_PRODUCT_ID_612A:
				UDIA_INFO("Microdia USB2.0 Webcam - Product ID 612A.\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_612A;
				set_contrast = NULL;	// sn9c325_set_contrast
				set_brightness = NULL;	// sn9c325_set_brightness
				set_gamma = NULL;	// sn9c325_set_gamma
				set_sharpness = NULL;	// sn9c325_set_sharpness
				return -ENODEV;

			default:
				return -ENODEV;
		}
	}
	// "Clone"-devices:
	else if (vendor_id == USB_VENDOR_ID_MICROSOFT) {
		switch (product_id) {
			case USB_UDIA_MICROSOFT_PRODUCT_ID_00F4:
				UDIA_INFO("Microdia PID 628F Compatible USB2.0 Webcam:\n          Microsoft LifeCam VX-6000 (045e:00f4)\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_628F;
				webcam_type = MICRODIA_SXGA;
				sensor_slave_address = 0x30;
				return -ENODEV;
			default:
				return -ENODEV;
		}
	}
	else if (vendor_id == USB_VENDOR_ID_CHICONY) {
		switch (product_id) {
			case USB_UDIA_CHICONY_PRODUCT_ID_A128:
				UDIA_INFO("Microdia PID 62BE Compatible USB2.0 Webcam:\n          Chicony Panda 7 (04f2:a128)\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_62BE;
				return -ENODEV;
			default:
				return -ENODEV;
		}
	}
	else if (vendor_id == USB_VENDOR_ID_TRUST) {
		switch (product_id) {
			case USB_UDIA_TRUST_PRODUCT_ID_013D:
				UDIA_INFO("Microdia PID 627B Compatible USB2.0 Webcam:\n          Trust WB-3600R (145f:013d)\n");
				UDIA_INFO("This device is not supported yet. Please contact microdia@googlegroups.com if you want to add support for this device\n");
				webcam_model = MICRODIA_627B;
				sensor_slave_address = 0x21;
				return -ENODEV;
			default:
				return -ENODEV;
		}
	}
	else
		return -ENODEV;

	// Allocate structure, initialize pointers, mutexes, etc. and link it to the usb_device
	dev = kzalloc(sizeof(struct usb_microdia), GFP_KERNEL);

	if (dev == NULL) {
		UDIA_ERROR("Out of memory !\n");
		return -ENOMEM;
	}

	// Init mutexes, spinlock, etc.
	init_MUTEX(&dev->mutex);
	spin_lock_init(&dev->spinlock);
	init_waitqueue_head(&dev->wait_frame);

	// Save pointers
	dev->webcam_model = webcam_model;
	dev->webcam_type = webcam_type;
	if (webcam_model == MICRODIA_6260)
		dev->frame_size_divisor = MICRODIA_FRAME_SIZE_DIVISOR_6260;
	else
		dev->frame_size_divisor = MICRODIA_FRAME_SIZE_DIVISOR_DEFAULT;
	dev->initialize = initialize;
	dev->stop_stream = stop_stream;
	dev->start_stream = start_stream;
	dev->set_contrast = set_contrast;
	dev->set_brightness = set_brightness;
	dev->set_gamma = set_gamma;
	dev->set_exposure = set_exposure;
	if (flip_detect) {
		UDIA_INFO("Rotate detection enabled\n");
		dev->flip_detect = flip_detect_f;
	}
	dev->set_hvflip = set_hvflip;
	dev->set_sharpness = set_sharpness;
	dev->set_auto_exposure = set_auto_exposure;
	dev->set_auto_whitebalance = set_auto_whitebalance;
	dev->udev = udev;
	dev->interface = interface;
	dev->sensor_slave_address = sensor_slave_address;
	dev->sensor_flags = sensor_flags;

	// Read the product release 
	dev->release = le16_to_cpu(udev->descriptor.bcdDevice);
	UDIA_INFO("Release: %04x\n", dev->release);

	// How many interfaces (1 or 3) ?
	bNumInterfaces = udev->config->desc.bNumInterfaces;
	UDIA_INFO("Number of interfaces : %d\n", bNumInterfaces);


	// Constructor
	dev->nbuffers = 2;
	dev->len_per_image = PAGE_ALIGN((1280 * 1024 * 4));


	// Switch on the camera (to detect size of buffers)
	dev_microdia_camera_on(dev);


	// Set up the endpoint information 
	// use only the first int-in and isoc-in endpoints
	// for the current alternate setting
	iface_desc = interface->cur_altsetting;

	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		if (!dev->int_in_endpointAddr
				&& ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
				&& ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_INT)) {
			// we found an interrupt in endpoint
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);

			dev->int_in_size = buffer_size;
			dev->int_in_endpointAddr = (endpoint->bEndpointAddress & 0xf);
		}

		if (!dev->isoc_in_endpointAddr
				&& ((endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN)
				&& ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_ISOC)) {
			// we found an isoc in endpoint
			buffer_size = le16_to_cpu(endpoint->wMaxPacketSize);

			dev->isoc_in_size = buffer_size;
			dev->isoc_in_endpointAddr = (endpoint->bEndpointAddress & 0xf);
		}
	}

	if (!(dev->int_in_endpointAddr && dev->isoc_in_endpointAddr)) {
		UDIA_ERROR("Could not find both int-in and isoc-in endpoints");

		kfree(dev);

		return -ENODEV;
	}


	// Switch off camera
	dev_microdia_camera_off(dev);

	// Initialize the video device
	dev->vdev = video_device_alloc();

	if (!dev->vdev) {
		kfree(dev);
		return -ENOMEM;
	}

	// Initialize the camera
	dev_microdia_initialize_device(dev);
	
	// Register the video device
	err = v4l_microdia_register_video_device(dev);

	if (err) {
		video_device_release(dev->vdev);	// Oops release video dev
		kfree(dev);
		return err;
	}

	// Create the entries in the sys filesystem
	microdia_create_sysfs_files(dev->vdev);

	// Save our data pointer in this interface device
	usb_set_intfdata(interface, dev);

	// Default settings video device
	usb_microdia_default_settings(dev);
	

	return 0;
}


/** 
 * @param interface 
 *
 * @brief This function is called when the device is disconnected
 *   or when the kernel module is unloaded.
 */
static void usb_microdia_disconnect(struct usb_interface *interface)
{
	struct usb_microdia *dev = usb_get_intfdata(interface);

	UDIA_INFO("Microdia USB2.0 Camera disconnected\n");

	usb_set_intfdata(interface, NULL);

	// We got unplugged; this is signalled by an EPIPE error code
	if (dev->vopen) {
		UDIA_INFO("Disconnected while webcam is in use !\n");
		dev->error_status = EPIPE;
	}

	// Alert waiting processes
	wake_up_interruptible(&dev->wait_frame);

	// Wait until device is closed
	while (dev->vopen)
		schedule();

	// Remove the entries in the sys filesystem
	microdia_remove_sysfs_files(dev->vdev);

	// Unregister the video device
	v4l_microdia_unregister_video_device(dev);
}


/**
 * @var usb_microdia_driver
 *
 * This variable contains some callback
 */
static struct usb_driver usb_microdia_driver = {
	.name = "usb_microdia_driver",
	.probe = usb_microdia_probe,
	.disconnect = usb_microdia_disconnect,
	.id_table = microdia_table,
};

module_param(fps, int, 0444);			/**< @brief Module parameter frames per second */
module_param(hflip, int, 0444);			/**< @brief Module parameter horizontal flip process */
module_param(vflip, int, 0444);			/**< @brief Module parameter vertical flip process */
module_param(flip_detect, int, 0444);		/**< @brief Module parameter flip detect */
module_param(auto_exposure, int, 0444);		/**< @brief Module parameter automatic exposure control */
module_param(auto_whitebalance, int, 0444);	/**< @brief Module parameter automatic whitebalance control */
module_param(brightness, int, 0444);		/**< @brief Module parameter brightness */
module_param(whiteness, int, 0444);		/**< @brief Module parameter whiteness */
module_param(contrast, int, 0444);		/**< @brief Module parameter contrast */
module_param(exposure, int, 0444);		/**< @brief Module parameter exposure */
module_param(sharpness, int, 0444);		/**< @brief Module parameter sharpness */


/** 
 * @returns 0 if all is OK
 *
 * @brief Initialize the driver.
 *
 * This function is called at first.
 * This function permits to define the default values from the command line.
 */
static int __init usb_microdia_init(void)
{
	int result;

	UDIA_INFO("Microdia USB2.0 webcam driver startup\n");

	debug_dir = debugfs_create_dir(debug_dir_name, NULL);
	if(debug_dir)
		debug_file = debugfs_create_u8(debug_file_name, S_IRUGO | S_IWUGO,
					debug_dir, &debug_level);

	if(fps < 10 || fps > 30) {
		UDIA_WARNING("Framerate out of bounds [10-30]! Defaulting to 25\n");
		fps = 25;
	}

	if(vflip != 0 && vflip != 1) {
		UDIA_WARNING("Vertical flip should be 0 or 1! Defaulting to 0\n");
		vflip = 0;
	}

	if(hflip != 0 && hflip != 1) {
		UDIA_WARNING("Horizontal flip should be 0 or 1! Defaulting to 0\n");
		hflip = 0;
	}

	if(sharpness < 0 || sharpness > 0x3f) {
		UDIA_WARNING("Sharpness should be 0 to 63 ! Defaulting to 31\n");
		sharpness = 0x1f;
	}

	if(auto_exposure != 0 && auto_exposure != 1) {
		UDIA_WARNING("Automatic exposure should be 0 or 1! Defaulting to 0\n");
		vflip = 0;
	}

	if(auto_whitebalance != 0 && auto_whitebalance != 1) {
		UDIA_WARNING("Automatic whitebalance should be 0 or 1! Defaulting to 0\n");
		hflip = 0;
	}

	// Register the driver with the USB subsystem
	result = usb_register(&usb_microdia_driver);

	if (result)
		UDIA_ERROR("usb_register failed ! Error number %d\n", result);

	UDIA_INFO(DRIVER_VERSION " : " DRIVER_DESC "\n");

	return result;
}


/** 
 * @brief Close the driver
 *
 * This function is called at last when you unload the driver.
 */
static void __exit usb_microdia_exit(void)
{
	UDIA_INFO("usb_microdia_exit: Microdia USB2.0 webcam driver shutdown\n");

	if(debug_file)
		debugfs_remove(debug_file);
	if(debug_dir)
		debugfs_remove(debug_dir);

	// Deregister this driver with the USB subsystem
	usb_deregister(&usb_microdia_driver);
}


module_init(usb_microdia_init);				/**< @brief Module initialize */
module_exit(usb_microdia_exit);				/**< @brief Module exit */


MODULE_PARM_DESC(fps, "Frames per second [10-30]");	/**< @brief Description of 'fps' parameter */
MODULE_PARM_DESC(hflip, "Horizontal image flip");	/**< @brief Description of 'hflip' parameter */
MODULE_PARM_DESC(vflip, "Vertical image flip");		/**< @brief Description of 'vflip' parameter */
MODULE_PARM_DESC(flip_detect, "Image flip detection");		/**< @brief Description of 'vflip_detect' parameter */
MODULE_PARM_DESC(auto_exposure, "Automatic exposure control");	/**< @brief Description of 'auto_exposure' parameter */
MODULE_PARM_DESC(auto_whitebalance, "Automatic whitebalance");	/**< @brief Description of 'auto_whitebalance' parameter */
MODULE_PARM_DESC(brightness, "Brightness setting");	/**< @brief Description of 'brightness' parameter */
MODULE_PARM_DESC(whiteness, "Whiteness setting");	/**< @brief Description of 'whiteness' parameter */
MODULE_PARM_DESC(exposure, "Exposure setting");		/**< @brief Description of 'exposure' parameter */
MODULE_PARM_DESC(contrast, "Contrast setting");		/**< @brief Description of 'contrast' parameter */
MODULE_PARM_DESC(sharpness, "Sharpness setting");	/**< @brief Description of 'sharpness' parameter */


MODULE_LICENSE("GPL");					/**< @brief Driver is under licence GPL */
MODULE_AUTHOR(DRIVER_AUTHOR);				/**< @brief Driver is written by Nicolas VIVIEN */
MODULE_DESCRIPTION(DRIVER_DESC);			/**< @brief Define the description of the driver */
MODULE_SUPPORTED_DEVICE(DRIVER_SUPPORT);		/**< @brief List of supported device */


